#################
## Construct B-splines basis, penalty matrix, and some other quantities for smooth.X, by extracting attributes from a mgcv gam fit
## The code below is used as part of the vagam function in the vagam_main.R script

## no. of total knots = no. of interior knots[1] + 2*degree + 1, but then -1 to remove the intercept term 
## no. of basis dimension = d_j = no. of interior knots[1]+ degree + 1, where degree of spline is typically set to 3 for cubic B-splines. But then we typically -1 to remove the intercept term. Then to ensure centering constraint is the satisfied, reduce by one to produce to d_j = no. of interior knots[1] + degree - 1 
##################
library(mgcv)

## Please supply smooth.X, which is the model matrix containing the covariates to be included in the smoothing component. For example
# source("gamsim_fun.R")
# choose.n <- 100
# normal_dat <- gamsim(n = choose.n, dist = "normal", extra.X = data.frame(intercept = rep(1,choose.n), treatment = rep(c(0,1), each = choose.n/2)), beta = c(-1, 0.5))
# smooth.X <- normal_dat[,2:5]

int.knots <- 5 ## K_j
ncov.smooth.X <- ncol(smooth.X) ## This is the number of covariates to smooth i.e., q
actual.int.knots <- rep(int.knots, ncov.smooth.X)

S <- basis.info <- vector("list", ncov.smooth.X) ## All the difference matrices
d <- numeric(ncov.smooth.X) ## rank of S_j
index.cov <- NULL ## Indexes which covariate each column in Z belongs to
Z <- NULL ## full set of B-spline basis functions 

for(k in 1:ncov.smooth.X) {
	get.basis <- smooth.construct(s(x0, bs = "ps", k = actual.int.knots[k]+3, m = c(2,1)), data = data.frame(x0 = smooth.X[,k]), knots = NULL) ## m = c(2,1) implies a cubic spline i.e., m_j = 3 and a first order difference penalty. 
	Z.k <- get.basis$X

	## Construct centering constraint
	transform.mat <- qr.Q(qr(as.matrix(colSums(Z.k))), complete = TRUE)[,-1] 

	## Apply centering constraint to Z
	Z.k <- Z.k %*% transform.mat
	index.cov <- c(index.cov, rep(k, ncol(Z.k)))
	colnames(Z.k) <- paste("smooth.X", k, 1:sum(index.cov == k), sep = "")
	Z <- cbind(Z, Z.k)

	## Apply centering constraint to difference matrix
	S[[k]] <- t(transform.mat) %*% get.basis$S[[1]] %*% transform.mat
	d[k] <- get.basis$rank
	get.basis$X <- NULL; 
	get.basis$transform.mat <- transform.mat; 
	basis.info[[k]] <- get.basis
	rm(transform.mat, get.basis)
	}
