###########################
## VAGAMs auxiliary functions
##############################

## Function for calculating VA logL
## d is rank of each element in S
## index.cov indexes which covariate each column in Z belongs to
calc.VAlogL <- function(y, Z, para.X, family, lambda, kappa, a, A, phi, S, d, index.cov, eta = NULL) 
{
	if(is.null(eta)) 
          eta <- para.X %*% kappa + Z %*% a
	if(family$family == "gaussian")
	{
		out <- sum(dnorm(y, mean = eta, sd = sqrt(phi), log = TRUE)) - (0.5/phi) * sum((Z %*% A) * Z) + 0.5 * determinant(A)$mod + 0.5 * sum(d * log(lambda[lambda > 0]))
	}
	if(family$family == "binomial")
	{
          ZAZ <- (Z %*% A) * Z
		out <- sum(dbinom(y, 1, p = binomial()$linkinv(eta + 0.5 * rowSums(ZAZ)), log = TRUE)) - (0.5/phi) * sum(ZAZ) + 0.5 * determinant(A)$mod + 0.5 * sum(d * log(lambda[lambda > 0]))
	}
	if(family$family == "binomiall")
	{
          ZAZ <- rowSums((Z %*% A) * Z)
		out <- sum(dbinom(y, 1, p = binomial()$linkinv(eta + 0.5 * ZAZ), log = TRUE)) - 0.5 * sum(y * ZAZ) + 0.5 * determinant(A)$mod + 0.5 * sum(d * log(lambda[lambda > 0]))
	}
	if(family$family == "poisson") 
	{
          ZAZ <- rowSums((Z %*% A) * Z)
		out <- sum(dpois(y, lambda = exp(eta + 0.5 * ZAZ), log = TRUE)) - 0.5 * sum(y * ZAZ) + 0.5 * determinant(A)$mod + 0.5 * sum(d * log(lambda[lambda > 0]))
	}

	for(k2 in 1:length(lambda)) 
	{
		out <- out - 0.5 * lambda[k2] * sum(diag(crossprod(S[[k2]], A[index.cov == k2, index.cov == k2]))) - 0.5 * lambda[k2] * t(a[index.cov == k2]) %*% S[[k2]] %*% a[index.cov == k2]
	}
	return(as.numeric(out))
}


## calculate information matrix via Louis method
info.valouis <- function(y, para.X, Z, kappa, phi = 1, lambda, a, A, S, d, index.cov, mc.samps = 2000, family = gaussian(), seed = NULL)
{
	if(!is.null(seed))
	{
		old <- .Random.seed
		on.exit( { .Random.seed <<- old } )	
		set.seed(seed)
	}
	
	diminfo <- ncol(para.X) + length(lambda) + as.numeric(family$family == "gaussian")
	n <- length(y)
	mc.betas <- rmvnorm(mc.samps, mean = a, sigma = A) 
	mc.betasTSmc.betas <- matrix(0, nrow = mc.samps, ncol = length(lambda))
	for(k in 1:length(lambda)) ## beta^T_j%*%S_j%*%beta_j, across the MC samples
	{
		mc.betasTSmc.betas[,k] <- rowSums((mc.betas[,index.cov==k] %*% S[[k]]) * mc.betas[,index.cov == k])
	}
	score.lambda <- matrix(0.5 * d/lambda, nrow = mc.samps, ncol = length(lambda), byrow = TRUE) - 0.5 * mc.betasTSmc.betas
	rm(mc.betasTSmc.betas)
	
	
	scorescoreT.out <- matrix(0, nrow = diminfo, ncol = diminfo) ## \sum\limits_{i=1}^n E{\partial \ell_{com}(y_i, beta) \partial \ell_{com}(y_i, beta)^T}
	neghess.out <- matrix(0, nrow = diminfo, ncol = diminfo) ## \sum\limits_{i=1}^n E{\partial^2 \ell_{com}(y_i, beta)}

	if(family$family == "gaussian")
	{
		for(i in 1:mc.samps)
		{
			res <- y - eta.para - as.vector(Z %*% mc.betas[i,])
			cw.score <- c(crossprod(para.X,res)/phi, -0.5 * n/phi + 0.5/phi^2 * sum(res^2))
			cw.score <- c(cw.score, score.lambda[i,])
			scorescoreT.out <- scorescoreT.out + tcrossprod(cw.score)
		}
		scorescoreT.out <- scorescoreT.out/mc.samps	
		res <- y - eta.para - Z %*% a
		neghess.out[1:ncol(para.X),1:ncol(para.X)] <- crossprod(para.X)/phi
		neghess.out[1:ncol(para.X),ncol(para.X)+1] <- crossprod(para.X, res)/phi^2
		neghess.out[ncol(para.X)+1,1:ncol(para.X)] <- t(neghess.out[1:ncol(para.X),ncol(para.X)+1])
		neghess.out[ncol(para.X)+1,ncol(para.X)+1] <- -0.5 * n/phi^2 + 1/phi^3 * (sum(res^2) + sum((Z %*% A) * Z))
		diag(neghess.out)[(ncol(para.X)+2):diminfo] <- 0.5 * d/lambda^2
	}
	if(family$family == "poisson")
	{
		for(i in 1:mc.samps) 
		{
			res <- y - exp(eta.para + as.vector(Z %*% mc.betas[i,]))
			cw.score <- c(crossprod(para.X, res), score.lambda[i,])
			scorescoreT.out <- scorescoreT.out + tcrossprod(cw.score)
		}
		scorescoreT.out <- scorescoreT.out/mc.samps	
		res <- exp(eta.para + Z %*% a + 0.5 * rowSums((Z %*% A) * Z))
		neghess.out[1:ncol(para.X),1:ncol(para.X)] <- t(para.X) %*% (para.X * as.vector(res))
		diag(neghess.out)[(ncol(para.X)+1):diminfo] <- 0.5 * d/lambda^2	
	}
	if(family$family == "binomial") 
	{
		all.eta <- eta.para + Z %*% a
		seq.a <- -Inf * (y == 0); seq.a[is.nan(seq.a)] <- 0
		seq.b <- Inf * (y == 1); seq.b[is.nan(seq.b)] <- 0
		
		for(i in 1:mc.samps) 
		{
			mc.u <- rtruncnorm(n, a = seq.a, b = seq.b, mean = all.eta)
			res <- (mc.u - eta.para -  as.vector(Z %*% mc.betas[i,]))
			cw.score <- c(t(para.X) %*% res, score.lambda[i,])
			scorescoreT.out <- scorescoreT.out + tcrossprod(cw.score)
		}
		scorescoreT.out <- scorescoreT.out/mc.samps	
		neghess.out[1:ncol(para.X),1:ncol(para.X)] <- crossprod(para.X)
		diag(neghess.out)[(ncol(para.X)+1):diminfo] <- 0.5 * d/lambda^2
	}
	if(family$family == "binomiall")
	{
		meanw <- numeric(n)
		for(i in 1:mc.samps) 
		{
			res <- binomial()$linkinv(eta.para + as.vector(Z %*% mc.betas[i,]))
			cw.score <- c(t(para.X) %*% (y-res), score.lambda[i,])
			scorescoreT.out <- scorescoreT.out + tcrossprod(cw.score)
			meanw <- meanw + binomial()$variance(res)
		}
		scorescoreT.out <- scorescoreT.out/mc.samps	
		meanw <- meanw/mc.samps
		neghess.out[1:ncol(para.X),1:ncol(para.X)] <- t(para.X) %*% (para.X * as.vector(meanw))
		diag(neghess.out)[(ncol(para.X)+1):diminfo] <- 0.5 * d/lambda^2	
	}

	rm(cw.score, score.lambda, res)
	get.info <- neghess.out - scorescoreT.out
	return(get.info)
}


# plot function for VGAM

plot.vagam <- function(x, n = 100, rug = TRUE, se = TRUE, xlim = NULL, ylim = NULL, xlab = NULL, ylab = NULL, main = NULL, select = NULL, ...)
{
	if(class(x) != "vagam")
	{
		stop("x should be an object of class vagam. Thanks!")
	}
	if(is.null(x$smooth.X))
	{
		stop("Fitted GAM must contain the data. Thanks!")
	}
  
	if(is.null(select)) sel.plots <- 1:ncol(x$smooth.X)
	if(!is.null(select)) sel.plots <- select
	for(k2 in sel.plots) 
	{
		calc.pred <- predict.vagam(x = x, new.smoothX = seq(min(x$smooth.X[,k2]), max(x$smooth.X[,k2]),
								length = 100), terms = k2)
		
		plot(calc.pred$new.smoothX, calc.pred$pred, type = "l", xlim = xlim, ylim = ylim, main = main,
			xlab = ifelse(is.null(xlab), colnames(x$smooth.X)[k2], xlab), ylab = ifelse(is.null(ylab), paste("Smooth of", colnames(x$smooth.X)[k2]), ylab), ...)
		if(se)
		{
			lines(calc.pred$new.smoothX, calc.pred$lower, lty = 2, ...)
			lines(calc.pred$new.smoothX, calc.pred$upper, lty = 2, ...)			
		}
		if(rug)
		{
			rug(x$smooth.X[,k2], ...)
		}
	}
}
	
	
## x: fitted VAGAM model
## new.smoothX: new smoothing covariates
## new.paraX: new paramatric covariates
predict.vagam <- function(x, new.smoothX, new.paraX = NULL, terms = NULL, alpha = 0.05, type = "link")
{
	if(class(x) != "vagam")
	{
		stop("x should be an object of class vagam. Thanks!")
	}
	if(alpha > 1 || alpha < 0)
	{
		stop("alpha should be a number between 0 and 1, for (1-alpha)% pointwise confidence intervals. Thanks!")
	}
	type <- match.arg(type, c("link", "response"))	
	if(!is.null(terms)) 
	{
		if(length(terms) != 1) 
			stop("terms should either be NULL, in which case the prediction is made on all terms, or should be a single number indicating which smoothing term is to be used for prediction. Thanks!")
		if(!is.vector(new.smoothX))
			stop("If terms is a single number indicating which smoothing term is to be used for prediction, then new.smoothX should be a vector corresponding to the covariate values for which predictions are to be calculated for that term. Thanks!")
	}
	if(is.null(terms))
	{
		if(ncol(new.smoothX) != length(x$no.knots))
			stop("If terms is NULL, in which case the prediction is made on all terms, then new.smoothX should be a matrix with the name number of columns as the number of smoothing covariates fitted in x. Thanks!")
	}
	if(!is.null(new.paraX))
	{
		if(any(apply(new.paraX, 2, function(x) all(x == 1))))
			stop("No intercept terms should be included in new.paraX, as this is included by default. Thanks!")
			new.paraX <- cbind(1, new.paraX) ## Intercept included by default
		if(is.null(x$para.stat))
			stop("If new.paraX is supplied, in which case the prediction involves parametric terms, then x should include the relevant standard errors i.e., x$para.stat. Thanks!")
	}

	
	if(!is.null(terms)) ## Predictions made for a single smoothing covariates
	{
		# calculate total variance for a smoothing covariate
		newZ <- Predict.matrix(x$basis.info[[terms]], data = data.frame(x0 = (new.smoothX)))
		newZ <- newZ %*% x$basis.info[[terms]]$transform.mat
		new.eta <- newZ %*% x$a[x$index.cov == terms]
		sub.A <- x$A[x$index.cov == terms, x$index.cov == terms]
		variance <- rowSums((newZ %*% sub.A) * newZ)
		
		# construct pointwise prediction interval based on normality assumption
		lower.bound <- new.eta - qnorm(alpha/2, lower = FALSE) * sqrt(variance)
		upper.bound <- new.eta + qnorm(alpha/2, lower = FALSE) * sqrt(variance)
	}
	
	
	if(is.null(terms)) ## Predictions made across the smoothing (and parametric if supplied) covariates. Note the predictions will account for the intercept ONLY if new.paraX is supplied
	{
		# calculate total variance for smoothing covariates
		newZ <- NULL
		for(k2 in 1:length(x$no.knots)) 
		{
			tmpZ <- Predict.matrix(x$basis.info[[k2]], data = data.frame(x0 = (new.smoothX[,k2])))
			newZ <- cbind(newZ, tmpZ %*% x$basis.info[[k2]]$transform.mat)
		}
		new.eta <- newZ %*% x$a
		variance <- rowSums((newZ %*% x$A) * newZ)

		if(!is.null(new.paraX))
		{
		new.eta <- new.eta + as.matrix(new.paraX)%*%x$kappa
		variance <- variance + rowSums((new.paraX %*% solve(x$obs.info)[1:length(x$kappa),1:length(x$kappa)]) * new.paraX)
		}
		
		# construct pointwise prediction interval based on normality assumption
		lower.bound <- new.eta - qnorm(alpha/2, lower = FALSE) * sqrt(variance)
		upper.bound <- new.eta + qnorm(alpha/2, lower = FALSE) * sqrt(variance)
	}
	if(type == "response")
	{
		new.eta <- x$family$linkinv(new.eta)
		lower.bound <- x$family$linkinv(lower.bound)
		upper.bound <- x$family$linkinv(upper.bound)
	}
	out <- data.frame(prediction = new.eta, se = sqrt(variance), lower.bound = lower.bound, upper.bound = upper.bound,
					new.smoothX = new.smoothX)
	if(!is.null(new.paraX))
		out$new.paraX <- new.paraX

	return(out)
}


## Iterative Weighted Least Squares for GLMS with a general ridge penalty
ridge.iwls <- function(x, y, family, penalty = NULL, initial.beta = NULL, max.steps = 1000, conv.eps = 1e-3, offset = NULL, gamma = 1)
{
	y <- as.vector(y)
	x <- as.matrix(x)
	p <- ncol(x)
	nobs <- nrow(x)
	if(is.null(penalty)) 
		penalty <- matrix(0, p, p)
	if(is.null(offset)) 
		offset <- numeric(length(y))
	if(!is.null(penalty))
	{
		if(nrow(penalty) != p || ncol(penalty) != p) 
			stop("penalty could a square matrix with dimension ncol(x) by ncol(x). Thanks!")
	}
	converged <- FALSE
	stop.at <- max.steps

	beta.mat <- matrix(0, nrow = max.steps, ncol = p)
	if(is.null(initial.beta)) 
		initial.beta <- rep(0.01, p)
	eta.new <- x %*% initial.beta + offset

	for(i in 1:max.steps)
	{
		beta.mat[i,] <- initial.beta
		mu.new <- family$linkinv(eta.new)
		d.new <- family$mu.eta(eta.new)
		v.new <- family$variance(mu.new)
		weights <- c(d.new/sqrt(v.new))
		x.star <- weights * x
		y.tilde.star <- weights * (eta.new - offset + (y - mu.new)/d.new)
		p.imat.new <- crossprod(x.star) + penalty
		# print(p.imat.new)
		inv.pimat.new <- chol2inv(chol(p.imat.new))		
		beta.new <- gamma * inv.pimat.new %*% crossprod(x.star, y.tilde.star) + (1 - gamma) * beta.mat[i,]
		
		if((sum(abs(beta.new - initial.beta))/sum(abs(initial.beta)) <= conv.eps)) 
		{
			converged <- TRUE
			stop.at <- i
			if(i < max.steps) 
				break;
		}
		else
		{
			initial.beta <- beta.new
			eta.new <- x %*% beta.new + offset
		}
		# print(c(beta.new))
	}
	
	out <- list(coefficients = as.vector(beta.new), family = family, converged = converged,
                stop.at = stop.at, linear.predictors = eta.new)

     if(extra) {
          Infmat <- inv.pimat.new %*% crossprod(x.star)
          tr.Infmat <- sum(diag(Infmat))
          out$tr.Inf <- tr.Infmat
          }
     return(out)
}



